#ifndef QTELLASUB_H
#define QTELLASUB_H

#include <string>
#include <vector>
#include <list>
#include <set>

class QMenuBar;
class Connections;
class SharedFile;
class UploadManager;
class DownloadManager;
class IncomingConnect;
class QPopupMenu;
class HostCache;
class InterruptedDownloads;
class MergeDownloads;
class QCloseEvent;

class QTimer;
class ExecThread;

#include "qthread.h"
#include <qapplication.h>
#include <qstring.h>
#include <qfont.h>
#include <qpalette.h>

#include "Qtella.h"
#include "HashTable.h"

#ifdef KDE
#include "kmenubar.h"
#include "kpopupmenu.h"
#endif

#if !defined(_WIN32)
using namespace std;
#endif

class QtellaSub : public Qtella
{ 
    Q_OBJECT

 public:
    QtellaSub( QApplication* parent = 0, const char* name = 0, WFlags fl = 0 );
    ~QtellaSub();

    #ifdef KDE
    KMenuBar*            menu;
    KPopupMenu*          _help;
    KPopupMenu*          _file;
    KPopupMenu*          _style;
    KPopupMenu*          _view;
    #endif

    #ifdef QT
    QMenuBar*            menu;
    QPopupMenu*          _help;
    QPopupMenu*          _file;
    QPopupMenu*          _style;
    QPopupMenu*          _view;
    #endif

    Connections*         connections;
    QTimer*              timer_connection;
    
    std::vector< std::pair<std::string, std::string> > _applications;
    std::vector<SharedFile*>  vSharedFiles;

    unsigned long             number_shared_files;
    double                    shared_size;

    static void               sighandler( int );
    void                      checkDirectories( const std::string& directory, std::vector<SharedFile*>* old );
    void                      removeFiles( const int size );
    void                      launchFile( const std::string& filename );
    void                      launchFileList( std::vector<std::string>& files );
    static QtellaSub*         getInstance();
    const void                search( const std::string& query);
    const void                setReportedIP( const std::string& ip );

    UploadManager*            upload_manager;
    DownloadManager*          download_manager;
    IncomingConnect*          _incoming_connect;
    HostCache*                _hostcache;
    InterruptedDownloads*     _interrupted_downloads;
    MergeDownloads*           _merge_downloads;

    HashTable                 _hash_table;
    std::set<unsigned long>   _sharedsize_set;
    std::string               _style_s;

    QPopupMenu*               _context_download;
    QPopupMenu*               _context_search;
    QPopupMenu*               _context_upload;
    QPopupMenu*               _context_shared;
    QPopupMenu*               _context_connect;

    QPalette                  _palette;
    QPalette                  _menu_palette;

    int                       _search_column_size;
    unsigned                  _current_port;
    std::string               _last_download_to_dir;
    bool                      _incoming_allowed;

    const int                 strtoint(const std::string& s);
    
 public slots:

    virtual void slotSharedFilesLaunch(QListViewItem* item);
    virtual void slotSharedFilesLaunch();
    virtual void slotAddApplication();
    virtual void slotRemoveApplication();
    virtual void slotEditApplication();
    virtual void slotChangeFinished();
    virtual void slotChangeInterrupted();
    virtual void slotChangeHostlist();
    virtual void slotChangeLogfile();
    virtual void slotSelectHostCacheLogFile();
    virtual void slotDeleteDownload();
    virtual void slotCloseSearch();
    virtual void slotLaunchInterruptedFile();
    virtual void slotDeleteInterruptedFile();
    virtual void slotSearchInterruptedFile();
    virtual void slotResumeInterruptedFile();
    virtual void slotAbortInterruptedFile();
    virtual void slotAddBlockedHost();
    virtual void slotDeleteBlockedHost();
    virtual void slotUploadBlockHost();
    virtual void slotSearchStopDownload();
    virtual void slotSearchRemoveDownload();
    virtual void slotSearchDeleteDownload();
    virtual void slotViewChanged(int);

    void slotConnect();
    void slotDisconnect();
    void check_connections();
    void slotInfo();

    virtual void slotDownloadDoubleClicked(QListViewItem* item);
    virtual void slotDownloadShowContext(QListViewItem*, const QPoint&, int);
    virtual void slotLaunchSelectedDownloadItem();
    virtual void slotShowSearchContext(QListViewItem*, const QPoint&, int);
    virtual void slotShowUploadContext(QListViewItem*, const QPoint&, int);
    virtual void slotShowSharedContext(QListViewItem*, const QPoint&, int);
    virtual void slotSharedProperties();
    virtual void slotChangeInterface(const QString& text);
    virtual void slotClickedAutoConnect(int, QListBoxItem*, const QPoint&);
    virtual void slotConnectContext(QListViewItem*, const QPoint&, int);
    virtual void slotInterruptedFileContext(QListViewItem*, const QPoint&, int);
    virtual void slotUpdateAutoConnect();
    virtual void slotRefreshSharedList();
    virtual void slotSearch();
    virtual void slotAddDirectory();
    virtual void slotRemoveDirectory();
    virtual void slotApplyConfig();
    virtual void slotStopSearch();
    virtual void slotCurrentChanged(QWidget*);
    virtual void slotAddHostAutoConnect();

    virtual void slotDownloadAll();
    virtual void slotDownloadTo();
    virtual void slotSearchDoubleClicked(QListViewItem*);

    virtual void slotAddAutoConnect();
    virtual void slotRemoveAutoConnect();
    virtual void slotAddHost();
    virtual void slotAbortUpload();
    virtual void slotRemoveUpload();
    virtual void slotDeleteSharedFile();
    virtual void slotClearInactiveUploads();
    virtual void slotReadFile();
    virtual void slotRenameSharedFile();
    virtual void slotGroupSize(int);

    virtual void slotHostCacheChangePort(int port);
    virtual void slotHostCacheChangeEnable();
    
    virtual void closeEvent(QCloseEvent *); /* system tray patch */

    void         slotStylePlatinum();
    void         slotStyleWindows();
    void         slotStyleMotif();
    void         slotStyleMotifPlus();
    void         slotStyleCDE();
    void         slotStyleSGI();
    void         slotStyleKDE();
    void         slotStyle(int style);
    void         slotChangeFont();
    void         slotConvertFilenames();
    void         slotResubmitSearch();
    void         slotIPUpdate();

 private:
    friend class Download;
    friend class Connections;
    friend class QueryHitEntry;

    //! Set to save the pointers of items of all searches.
    /*!
      This set is used by the class Download to check whether
      the pointer of a download is still valid to update the
      status of the ball.
    */
    std::set<QListViewItem*>  _search_item_set;

    //! Sends the command to the DownloadManager to download a file.
    /*!
      This method is called by: <br>
      \sa slotSearchDoubleClicked(QListViewItem *item),
      \sa slotDownloadAll(),
      \sa slotDownloadTo()

      \param item Selected item in the search tab.

      \param dir Destination directory for the completed download. This
      directory is only valid during the current session. If Qtella crashes
      and later the download is resumed from the interrupted download list
      Qtella will save the complete file into the directory specified in the
      configuration tab. By default this directory is empty and the download
      is saved into the directory specified in the configuration tab.
    */
    void                      downloadItem(QListViewItem *item, const std::string& dir = "");

    void                      readConfiguration();
    void                      readHostList();

    std::string               _pref_if;
    std::string               _last_addr;

    std::string               _binary;
    std::string               conffile;
    int                       search_tabs;
    bool                      _read_config;
    bool                      _show_delete_dialog;

    QMutex                    mutex_shared_files;
    QFont                     _current_font;
    static QtellaSub*         _instance;
    std::string               _prev_version;
    QTimer*                   _ip_update_timer;
};

#endif // QTELLASUB_H
