#include "MP3property.h"
#include "StringManipulation.h"

#include <strstream>
#include <iostream>
#include <qfile.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qcombobox.h>

static int mpeg1Bitrates[] = { 0, 32, 40, 48, 56, 64, 80, 96, 112,
                               128, 160, 192, 224, 256, 320 };
static int mpeg2Bitrates[] = { 0, 8, 16, 24, 32, 40, 48, 56, 64,
                               80, 96, 112, 128, 144, 160 };
static int mpeg1SampleRates[] = { 44100, 48000, 32000 };
static int mpeg2SampleRates[] = { 22050, 24000, 16000 };
	
static std::string bitrate(std::string header)
{
   int id, br;

   id = ((header.data()[1] & 0x8) >> 3);
   br = ((header.data()[2] & 0xF0) >> 4);

   std::strstream str;
   // TODO: Add range checking
   str << (id ? mpeg1Bitrates[br] : mpeg2Bitrates[br]) << std::ends;
   return std::string( str.str() );
}

static std::string samplerate(std::string header)
{
   int id, sr;

   id = (header.data()[1] & 0x8) >> 3;
   sr = (header.data()[2] >> 2) & 0x3;

   std::strstream str;
   // TODO: Add range checking
   str << ( id ? mpeg1SampleRates[sr] : mpeg2SampleRates[sr] ) << std::ends;
   return std::string( str.str() );
}

static std::string channelmode(std::string header)
{
   switch ((header.data()[3] & 0xc0) >> 6) {
     case 0: return "Stereo";
     case 1: return "Joint stereo";
     case 2: return "Dual mono";
     case 3: return "Mono";
   }
   return "???";
}

static std::string mpeg_ver(std::string header)
{
   std::strstream str;
   str << (!((header.data()[1] & 0x8) >> 3) + 1) << std::ends;

   return std::string( str.str() );
}

static std::string mpeg_layer( std::string header )
{
  switch( (header.data()[1] & 0x6) >> 1 ) {
    case 1: return "3";
    case 2: return "2";
    case 3: return "1";
    default: return "?";
  }
  return "X";
}

static int padding(const char *header)
{
   return (header[2] >> 1) & 0x1;
}

static std::string copyright( std::string header )
{
   if( header.data()[3] & 0x8 >> 1 ) return "Yes";

   return "No";
}

static std::string original( std::string header )
{
   if( header.data()[3] & 0x4 >> 1 ) return "Yes";

   return "No";
}

static std::string crcprot( std::string header )
{
  if( header.data()[1] & 1 ) return "Yes";

  return "No";
}

/* 
 *  Constructs a MP3property which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
MP3property::MP3property( QWidget* parent,  const char* name, bool modal, WFlags fl )
    : BMP3Property( parent, name, modal, fl )
{
}

/*  
 *  Destroys the object and frees any allocated resources
 */
MP3property::~MP3property()
{
    // no need to delete child widgets, Qt does it all for us
}


void MP3property::setFile(std::string directory, std::string filename)
{
  _directory.assign(directory);
  _filename.assign(filename);

  std::string c(directory + "/" + filename);

  ui_textlabel_directory->setText(directory.c_str());
  ui_textlabel_filename->setText(filename.c_str());
  
  QFile f(c.c_str());
  std::strstream str;
  str << f.size() << std::ends;
  ui_textlabel_size->setText( StringManipulation::insertDelimiter(std::string(str.str()), ',', 3).c_str() );
  str.freeze(false);

  ui_textlabel_type->setText("Audio");

  if( f.open(IO_Raw | IO_ReadOnly) )
    {
      unsigned char buffer[128];
      
      int rd = f.readBlock((char *)buffer, 10);

      if( rd < 10 ) return;

      std::string s((char *)buffer, 10);

      // For a complete infomation on the MPEG Audio Layer I/II/III frame header see:
      // http://www.mp3-tech.org/programmer/frame_header.html

      if( buffer[0] == 0xff && buffer[1] & 0xf0 )	// YES, exact 0xfff nothing more!
	{
	  ui_textlabel_version->setText( mpeg_ver(s).c_str() );
	  ui_textlabel_layer->setText( mpeg_layer(s).c_str() );
	  ui_textlabel_bitrate->setText( bitrate(s).c_str() );
	  ui_textlabel_sampling->setText( samplerate(s).c_str() );
	  ui_textlabel_channel->setText( channelmode(s).c_str() );
	  ui_textlabel_copyright->setText( copyright(s).c_str() );
	  ui_textlabel_original->setText( original(s).c_str() );
	  ui_textlabel_crc->setText( crcprot(s).c_str() );
	}

      f.at( f.size() - 128 );
      rd = f.readBlock((char *)buffer, 128);

      f.close();

      if( rd != 128 ) return;

      s.insert( 0, (const char *)buffer, 128 );

      if( s.substr(0, 3) == std::string("TAG") )
	{
	  unsigned int track = s[126];
	  unsigned int genre = s[127];

	  if( genre > 80 ) genre = 80;

 	  ui_lineedit_title->setText(removeSpace(s.substr(3, 30)).c_str());
 	  ui_lineedit_artist->setText(removeSpace(s.substr(33, 30)).c_str());
 	  ui_lineedit_album->setText(removeSpace(s.substr(63, 30)).c_str());
 	  ui_lineedit_year->setText(removeSpace(s.substr(93, 4)).c_str());
 	  ui_lineedit_comment->setText(removeSpace(s.substr(97, 30)).c_str());
 	  ui_spinbox_track->setValue(track);
 	  ui_combobox_genre->setCurrentItem(genre);
	}

    }
}

const std::string MP3property::removeSpace(const std::string s)
{
  return s.substr(0, s.find_last_not_of(" \0") + 1);
}
