#include "Gnutella.h"
#include "Servent.h"
#include "Connections.h"
#include "QtellaSub.h"
#include "Logging.h"

#include "Ping.h"
#include "Host.h"
#include "Address.h"
#include "Message.h"
#include "PongCache.h"
#include "Bye.h"

#include <string>
#include <vector>
#include <strstream>

#include <qsocket.h>
#include <qlistview.h>
#include <qspinbox.h>
#include <qhostaddress.h>

#undef _DEBUG

Servent::Servent(Connections* parent, std::string id, QListViewItem *item, QSocket* socket, const std::string user_agent, const std::string handshake)
  : _id(id), _item(item), _timeout(0), _input(0), _output(0), _state(Idle),
    _parent(parent), _direction(0), _last_ping_send(0), _bytes_in(0), _bytes_out(0), 
    _bandwidth_in(0), _bandwidth_out(0), _queries(0),
    _timeout_counter(0), _bandwidth_counter(0), _ping_counter(0), _user_agent(user_agent),
    _header(handshake)
{
  _bye = 200;

  if( _user_agent.empty() )
    _version = "0.4";
  else
    _version = "0.6";

  _last_input = time(NULL);
  _start = time(NULL);

  _connect_str = CONNECT_STR;
  _response_str = RESPONSE_STR;

  if(socket) _socket = socket; else _socket = new QSocket();

  QObject::connect(_socket, SIGNAL(error(int)), SLOT(slotError(int)));
  QObject::connect(_socket, SIGNAL(connected()), SLOT(slotConnect()));
  QObject::connect(_socket, SIGNAL(readyRead()), SLOT(slotReadyRead()));
  QObject::connect(_socket, SIGNAL(connectionClosed()), SLOT(slotClosed()));
  QObject::connect(_socket, SIGNAL(delayedCloseFinished()), SLOT(slotDelayedClosed()));

  if(socket)
    if(_socket->state() == QSocket::Connection)
      {
	if( _version == "0.4" ) 
	  sendMessageDirect(_response_str);

	Ping p(1, 0, _id);
	sendMessageDirect(p._data);  // send initial ping
	_parent->_statistic.out_ping_n++;
	_parent->_statistic.out_ping_b += p._data.size();
	setState(Connected);
      }
    else
      setState(Closed);

  _servant_number = Servent::getNextServentNumber();

  _timeout_timer = new QTimer(this);
  QObject::connect(_timeout_timer, SIGNAL(timeout()), SLOT(timeout()));
  _timeout_timer->start(1000);
}


Servent::~Servent()
{
  std::strstream str;
  str << "Destructor called for " << _address.toString() << ". Servent: <" << _user_agent 
      << "> Buffer size: " << _buffer.size() << " Tx: " << _output
      << " Rx: " << _input << std::ends;
  Logging::getInstance()->log( Logging::Info, str.str() );
  str.freeze(false);

  PongCache::getInstance()->removePong(_servant_number);
  if(_item) delete _item;
  if(_timeout_timer) delete _timeout_timer;

  QObject::disconnect(_socket, 0, 0, 0);
  if(_socket) 
    {
      Bye bye( _bye );
      sendMessageDirect( bye.toString() );

      delete _socket;
    }
}


void Servent::setByeCode( const int code )
{
  _bye = code;
}


const int Servent::getBufferSize()
{
  return _buffer.size();
}


int Servent::getNextServentNumber()
{
  static unsigned int number = 1;
  ++number;
  if(!number) number = 1;
  return number;
}


void Servent::connect(Address address)
{
  _address = address;
  _strip.erase();
  _strip = address.strIP();

  setState(Connecting);
  _timeout = 0;
  _socket->connectToHost(_strip.c_str(), address.port());
}



void Servent::slotDelayedClosed()
{
  setState(Closed);
}


void Servent::close()
{
  Bye bye( _bye );
  sendMessageDirect( bye.toString() );

  _socket->close();
  setState(Closed);
}



int Servent::state()
{
  return _state;
}



void Servent::setState(int state)
{
  if( (state == Closed) && (!_message_buffer.empty()) )
    _state = Closing;
  else
    _state = state;
}



void Servent::slotError(int err)
{
  slotReadyRead();

  std::strstream str;
  str << "Connection error. Servent: <" << _user_agent 
      << "> Buffer size: " << _buffer.size() << " Tx: " << _output 
      << " Rx: " << _input << std::ends;
  Logging::getInstance()->log( Logging::Error, str.str() );
  str.freeze(false);

  if(_socket->state() != QSocket::Idle) _socket->close();
  setState(Error);
  _timeout = 0;
}



void Servent::writeConnectString()
{
  _timeout = 0;

  QHostAddress ha = _socket->peerAddress();
  
  // protocol v0.6 implementation, first connect string
  std::string connect_string("GNUTELLA CONNECT/0.6\r\nUser-Agent: Qtella ");
  connect_string += std::string(VERSION) + "\r\n";
  connect_string += std::string("Remote-IP: ") + std::string(ha.toString());
  connect_string += std::string("\r\n\r\n");
  sendMessageDirect(connect_string);

  // old protocol v0.4 implementation
  // sendMessageDirect(_connect_str);
}



void Servent::slotConnect()
{
  setState(Waiting);
  writeConnectString();
}



void Servent::slotClosed()
{
  slotReadyRead();

  std::strstream str;
  str << "Connection closed by foreign host. Servent: <" << _user_agent 
      << "> Buffer size: " << _buffer.size() << " Tx: " << _output
      << " Rx: " << _input << std::ends;
  Logging::getInstance()->log( Logging::Info, str.str() );
  str.freeze(false);

  setState(Closed);
}



void Servent::sendMessageDirect( const std::string& buffer )
{
  if(_socket->state() == QSocket::Connection)
    {
      _output++;
      _bytes_out += buffer.size();
      _socket->writeBlock(buffer.data(), buffer.size());
      _socket->flush();
    }
}



void Servent::sendMessage(std::string& buffer, bool cache)
{
  if( ((_send_buffer.size() < 512) && (_socket->bytesToWrite() < 1024))|| !cache )
    {
      _output++;
      _send_buffer.append(buffer);
    }
}


void Servent::limitVectors()
{
  while(_id_set.size() > 2000) _id_set.erase(_id_set.begin());
}


int Servent::getDirection()
{
  return _direction;
}


void Servent::setDirection(int i)
{
  _direction = i;
}


void Servent::slotReadyRead()
{
#ifdef _DEBUG
  qDebug("entering Servent::slotReadyRead");
#endif

  char   cbuff[4096+10];

  if( ! _socket->bytesAvailable() )
    return;

  _last_input = time(NULL);

  int bytes = 0;
  do
    {
      int rd = _socket->readBlock(cbuff, 4096);
      bytes += rd;
      _buffer.append(cbuff, rd);
    }
  while( _socket->bytesAvailable() );

#ifdef _DEBUG
  qDebug("Servent: %d read %d", _servant_number, bytes);
#endif

  if( (state() == Waiting) && (_buffer.find("\r\n\r\n") != std::string::npos) )
    {
      _header = _buffer.substr(0, _buffer.find("\r\n\r\n"));

      std::string b = _buffer;

      if( _buffer.substr(0, 9) != "GNUTELLA/" )
	{
#ifdef _DEBUG
	  qDebug("Servent error");
#endif
	  slotError(-1);
	  return;
	}

      _buffer.erase(0, 9);

      // get version
      
      std::string::size_type idx = _buffer.find(" ");
      if( idx == std::string::npos )
	{
#ifdef _DEBUG
	  qDebug("Servent error");
#endif
	  slotError(-1);
	  return;
	}

      _version = _buffer.substr(0, idx);

      _buffer.erase(0, idx + 1); // delete version number with space from buffer

      // get status code

      idx = _buffer.find(" ");
      if( idx == std::string::npos )
	{
#ifdef _DEBUG
	  qDebug("Servent error");
#endif
	  slotError(-1);
	  return;
	}

      std::string status = _buffer.substr(0, idx);
      _buffer.erase(0, idx + 1); // delete status code with space from buffer

      // get Remote-IP field
      idx = _buffer.find("Remote-IP:");
      if( idx != std::string::npos )
	{
	  idx = _buffer.find_first_of("0123456789", idx);
	  std::string::size_type idx2 = _buffer.find_first_not_of("0123456789.", idx);
	  if( idx2 != std::string::npos )
	    {
	      std::string ip = _buffer.substr( idx, idx2 - idx );
#ifdef _DEBUG
	      qDebug( "Found my IP: <%s>", ip.c_str() );
#endif
	      QtellaSub::getInstance()->setReportedIP( ip );
	    }
	}

      // get user agent information

      idx = _buffer.find("User-Agent:");
      if( idx != std::string::npos )
	{
	  _buffer.erase(0, idx + 11);
	  idx = _buffer.find_first_not_of(" ");
	  _buffer.erase(0, idx);
	  idx = _buffer.find("\r\n");
	  _user_agent = _buffer.substr(0, idx);
	}

      if( status == "503")
	{
#ifdef _DEBUG
	  qDebug("Servent: Busy <%s>", _buffer.c_str());
#endif
	  slotError(-1);
	  return;
	}

      if( status != "200")
	{
#ifdef _DEBUG
	  qDebug( "Servent: unknown status <%s>" );
#endif
	  std::strstream str;
	  str << _address.toString() << " Status: <" << status << ">" << " Servent: <" << _user_agent 
	      << "> Buffer size: " << _buffer.size() << " Tx: " << _output
	      << " Rx: " << _input << std::ends;
	  Logging::getInstance()->log( Logging::Error, str.str() );
	  str.freeze(false);

#ifdef _DEBUG
	  qDebug("Servent error");
#endif
	  slotError(-1);
	  return;
	}

      // status code is 200 ok

      idx = _buffer.find("\r\n\r\n"); // find end of header
      _buffer.erase(0, idx + 4);      // delete header from buffer

      setState(Connected);

      std::string r("GNUTELLA/0.6 200 OK\r\n\r\n");
      sendMessageDirect(r);

      int ttl = _parent->_parent->ui_spinbox_ttl->value();
      Ping p(1, 0, _id);
      sendMessageDirect(p._data);  // send initial ping
      _parent->_statistic.out_ping_n++;
      _parent->_statistic.out_ping_b += p._data.size();

#ifdef _DEBUG
      qDebug("Servent ok");
#endif
      return;
    }

  else

    while( _buffer.size() >= HEADER_SIZE )
      {
	Q_UINT32 payload  = Message::str_uint32(_buffer.substr(19, 4).data());

	if(payload > 20000)
	  {
	    _buffer.erase();
#ifdef _DEBUG
	    qDebug("Servent error (payload is big)");
#endif
	    slotError(-1);
	    return;
	  }
	
	if( _buffer.size() >= (HEADER_SIZE + payload) )
	  {
	    _input++;

	    if(_message_buffer.size() < 30)
	      _message_buffer.push_back(_buffer.substr(0, HEADER_SIZE + payload));
	    
	    _buffer.erase(0, HEADER_SIZE + payload);
	    _bytes_in += HEADER_SIZE + payload;
	  }
	else
	  return;
      }
#ifdef _DEBUG
  qDebug("exit Servent::slotReadyRead");
#endif
}


void Servent::timeout()
{
  _timeout_counter++;

  if(_timeout_counter >= 5)
    if( (time(NULL) - _last_input) >= 120 ) 
      {
	close();
	_timeout_counter = 0;
      }


  if( (_socket->state() == QSocket::Connection) && (state() == Connected) &&
      (_send_buffer.size() > 0) )
    {
      _bytes_out += _send_buffer.size();
      _socket->writeBlock(_send_buffer.data(), _send_buffer.size());
      _send_buffer.erase();
    }


  _bandwidth_counter++;

  if(_bandwidth_counter >= 2)
    {
      _bandwidth_in = static_cast<double>(_bytes_in) / static_cast<double>(2048);
      _bandwidth_out = static_cast<double>(_bytes_out) / static_cast<double>(2048);
      _bytes_in = 0;
      _bytes_out = 0;
      _bandwidth_counter = 0;
    }


  _ping_counter++;

  if(_ping_counter >= 3)
    {
      if( !_last_accepted_ping_id.empty() )
	if( ((static_cast<unsigned char>(_last_accepted_ping_id[8]) == 0xff) && 
	     (_last_accepted_ping_id[15] == 1)) ||
	    ((time(NULL) - _last_ping_send >= 30) && (state() == Connected)) )
	  {
	    int ttl = _parent->_parent->ui_spinbox_ttl->value();
	    Ping p(ttl, 0, _id);
	    sendMessage(p._data, false);
	    
	    _parent->_statistic.out_ping_n++;
	    _parent->_statistic.out_ping_b += p._data.size();
	    
	    _last_ping_send = time(NULL);
	  }
      _ping_counter = 0;
    }

  
  _queries = 0;
}


long Servent::uptime()
{
  return (time(NULL) - _start);
}
